# For use w/ SeqGeq
options(warn = - 1)

#### Places to Regex Variables
outPath = "SG_OUTPUT_DIRECTORY_PATH"
RunID = "SG_RUNID"
plotGenes = SG_PLOT_GENES
PCANUM = SG_PCA
qvalue = SG_QVAL_THRESHOLD
NumCells = SG_NUM_CELLS_EXPRESSION_FILTER
VPBool = TRUE

# Check if Monocle is installed, if not then try and install
# Install function for packages
loadPackages <-function(...) {
    libs<-unlist(list(...))
    req<-unlist(lapply(libs,require,character.only=TRUE))
    need<-libs[req==FALSE]
    if(length(need)>0){
        tryCatch(
        {
            install.packages(need)
        },
        error = function(cond) {
            install.packages(need, repos = "http://cran.us.r-project.org")
        }

        )
        lapply(need,require,character.only=TRUE)

    }
}

loadBioc<-function(...) {
    libs<-unlist(list(...))
    req<-unlist(lapply(libs,require,character.only=TRUE))
    need<-libs[req==FALSE]
    if(length(need)>0){
        BiocManager::install(need)
        lapply(need,require,character.only=TRUE)
    }
}
# check for packages then install if needed.
loadPackages(c('pheatmap', 'devtools', 'BiocManager', 'Matrix.utils', 'data.table', 'dplyr', 'ggplot2', 'tibble', 'spdep'))
loadBioc(c('BiocGenerics', 'DelayedArray', 'DelayedMatrixStats', 'limma', 'S4Vectors', 'SingleCellExperiment', 'SummarizedExperiment', 'batchelor', 'Matrix.utils', 'HDF5Array'))


tryCatch({
    pdf(NUL)
},
error = function(cond) {
    writeLines("Failed to open PDF maker.")
})

# if(!require('leidenbase',character.only=TRUE)){
#     suppressMessages(suppressWarnings(devtools::install_github('cole-trapnell-lab/leidenbase')))
# } else {
#     writeLines("Start loading leidenbase...")
#     library('leidenbase')
# }

if(!require('monocle3',character.only=TRUE)){
    writeLines(" Monocle3 is not installed, installing...")
    suppressMessages(suppressWarnings(devtools::install_github('cole-trapnell-lab/monocle3')))
} else {
    writeLines("Start loading Monocle3...")
    library('monocle3')
}

writeLines(" Importing CSV from SeqGeq")

# Read CSV from SeqGeq
importedCSVfile <- read.table(file= "SG_SAMPLE_DATA_PATH",
sep = ",", row.names = 1, header = TRUE, check.names = FALSE)
importedCSVfile = importedCSVfile[, colSums(importedCSVfile != 0) > 0]
colnames(importedCSVfile) <- paste("AppID", colnames(importedCSVfile),sep="_")


writeLines(" Read in file")
writeLines(" Start: Writting Monocle SCE object")

writeLines(" Clean data matrix..")
expressionMatrix <- importedCSVfile
colnames(expressionMatrix) <- NULL
cellAnnotations <- data.frame(sample = colnames(importedCSVfile), row.names = colnames(importedCSVfile))
colnames(cellAnnotations) <- c("Sample")

writeLines(" Prepare Monocle Objects..")
geneAnnotations <- data.frame(gene_short_name = rownames(importedCSVfile), row.names =  rownames(importedCSVfile))
# colnames(geneAnnotations) <- c("Genes")

## Step 1: create monocle3 object
writeLines("Write Monocle CellDataSet object..")
cds <- new_cell_data_set(as(as.matrix(expressionMatrix),'sparseMatrix'),
cell_metadata = cellAnnotations,
gene_metadata = geneAnnotations)

writeLines(" Created Monocle object")

writeLines(" Estimate Size Factors")
# Not working in m3
# cds <- estimate_size_factors(cds)
# cds <- estimateDispersions(cds)

#cds <- align_cds(cds, alignment_group = "batch")

## Step 2: Normalize & pre-Process to project the data onto the top principal components
cds <- preprocess_cds(cds, method = 'PCA', norm_method = "log", num_dim = PCANUM)
VarianceExplain <- plot_pc_variance_explained(cds)

ggplot2::ggsave(filename = paste0(outPath,"/VarianceExplain.png"), plot = VarianceExplain, width = 6, height = 6, units = "in", device = "png")


###################################
## Optional But needed for SeqGeq##
###################################
# Remove batch effects
# Need to add a column to colData that encodes 'SampleID' from SeqGeq concat
# cds <- align_cds(cds, alignment_group = "batch")
###################################
## Optional But needed for SeqGeq##
###################################


## Step 3: Reduce the dimensions using UMAP
cds <- reduce_dimension(cds = cds,
    max_components = PCANUM,
    reduction_method = "UMAP",
    umap.metric = "cosine",
    umap.min_dist = 0.1,
    umap.n_neighbors = 15L,
    umap.nn_method = "annoy"
)

## Step 4: Cluster the cells
cds <- cluster_cells(
    cds = cds,
    reduction_method = "UMAP",
    k = 20L,
    louvain_iter = 1L,
    partition_qval = qvalue,
    weight = FALSE,
    resolution = c(10L ^ seq(from = -6L, to = -1L)),
    verbose = TRUE
)


partitionPlot <- plot_cells(cds, color_cells_by="partition", group_cells_by="partition")
ggplot2::ggsave(filename = paste0(outPath,"/PartitionPlot.png"), plot = partitionPlot, width = 6, height = 6, units = "in", device = "png")


## Step 5: Learn a graph. Trajectory Analysis part 1
cds <- learn_graph(cds, use_partition = TRUE)

## Step 6: Order cells

# a helper function to identify the root principal points:
get_earliest_principal_node <- function(cds){
#cell_ids <- which(colData(cds)[, "Time"] == time_bin)
cell_ids <- colData(cds)[, 1]


closest_vertex <-
cds@principal_graph_aux[["UMAP"]]$pr_graph_cell_proj_closest_vertex
closest_vertex <- as.matrix(closest_vertex[colnames(cds), ])
root_pr_nodes <-
    igraph::V(principal_graph(cds)[["UMAP"]])$name[as.numeric(names
    (which.max(table(closest_vertex[cell_ids,]))))]

root_pr_nodes
}

#Could have user choose between UMAP and tSNE
cds <- order_cells(cds, root_pr_nodes=get_earliest_principal_node(cds), reduction_method = "UMAP", verbose = TRUE)

monocle_RootPlot <- plot_cells(cds,
    color_cells_by = "pseudotime",
    label_cell_groups=FALSE,
    label_leaves=FALSE,
    label_branch_points=FALSE,
    graph_label_size=1.5)

ggplot2::ggsave(filename = paste0(outPath,"/RootNode.png"), plot = monocle_RootPlot, width = 6, height = 6, units = "in", device = "png")

clusterPlot <- plot_cells(cds, color_cells_by="partition", group_cells_by="partition")
ggplot2::ggsave(filename = paste0(outPath,"/ClusterPlot.png"), plot = clusterPlot, width = 6, height = 6, units = "in", device = "png")


## Subset Cells by Branch ##
#######################################
#######################################

## cds_sub <- choose_graph_segments(cds)




## Automated Annotation with Garnett may be possible but seems tricky to implement. ##
#######################################
#######################################


## Differential Expression Analysis ##
#######################################
#######################################

## Maker Gene Expression by each cluster
marker_test_res <- top_markers(cds, group_cells_by="cluster",  reference_cells=1000, cores=8)
marker_test_res <- dplyr::filter(marker_test_res, marker_test_res$marker_test_q_value < qvalue)

top_specific_markers <- marker_test_res %>%
filter(fraction_expressing >= NumCells) %>%
group_by(cell_group) %>%
top_n(5, pseudo_R2)


writeLines(" Writting DE Genes for each cluster.")
DEGfile <- paste0(outPath, "/DEG_Clusters_", RunID, ".csv")
DEG <- apply(marker_test_res,2,as.character)

write.table("[GeneSetCollection]", file = DEGfile, append = F, quote = F, row.names = F, col.names = F)
write.table("$SchemaVersion,1", file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
write.table(paste0("$GeneSetCollectionName,", "Monocle_Genes_",RunID), file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
write.table("", file = DEGfile, append = T, quote = F, row.names = F, col.names = F)

i=1

for (ClustNum in unique(marker_test_res$cell_group)){
    write.table("[GeneSet]", file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
    write.table(paste0("$GeneSetName,","Genes_",ClustNum), file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
    write.table("$GeneSetDescription,MadeInMonocle3", file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
    write.table(" ", file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
    write.table("[Genes], q_value", file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
    write.table(DEG[DEG[,"cell_group"] == i, c("gene_short_name", "marker_test_q_value")], file = DEGfile, sep = ',', append = T, quote = F, row.names = F, col.names = F)
    i = i+1
}



#Plot the genes in a heatmap
top_specific_marker_ids <- unique(top_specific_markers %>% pull(gene_id))

GeneClustPlot <- plot_genes_by_group(cds,
                            top_specific_marker_ids,
                            group_cells_by="cluster",
                            ordering_type="maximal_on_diag",
                            max.size=3)

ggplot2::ggsave(filename = paste0(outPath,"/TopsGenesHeatmap.png"), plot = GeneClustPlot, width = 12, height = 12, units = "in", device = "png")


#Plot the genes on the pseudotime
PseudotimeGenesPlot <- plot_cells(cds, genes=top_specific_marker_ids,
                            show_trajectory_graph=FALSE,
                            label_cell_groups=FALSE,
                            label_leaves=FALSE)

plotname = paste0(outPath, "PseudotimeGenesPlot.png")
ggplot2::ggsave(filename = plotname, plot = PseudotimeGenesPlot, width = 12, height = 12, units = "in", device = "png")


#Plot each individual gene across pseudotime
if (plotGenes) {

    for (r in unique(marker_test_res$cell_group)){
        DEG2 <- subset(marker_test_res, cell_group==r)
        DEG2 <- DEG2[order(DEG2$specificity, decreasing = TRUE),]
        nGenes <- nrow(DEG2)
        if (nGenes >5){
            DEG2 <- DEG2[1:5,]
        }

        if (nGenes >0){
            plotHM = plot_genes_in_pseudotime(cds[rowData(cds)$gene_short_name %in% DEG2$gene_short_name,],
            #color_cells_by="pseudotime",
            min_expr =0.5)
        }
        TimeImageName <- paste0(outPath, "Pseudotime_Genes_", r, ".png")
        ggplot2::ggsave(filename = TimeImageName, plot = plotHM, width = 6, height = 6, units = "in", device = "png")
    }

}

    # gene_fits <- fit_models(cds, model_formula_str = "~cluster")
    # fit_coefs <- coefficient_table(gene_fits)
    # #fit_coefs %>% filter(p_value < 0.0005)
    # sig_fit_coefs <- fit_coefs %>% select(term, gene_short_name, num_cells_expressed, gene_id, status, estimate, std_err, test_val, p_value, q_value)
    # sig_fit_coefs <- dplyr::filter(sig_fit_coefs, sig_fit_coefs$p_value < 0.0005)
    #
    # writeLines(" Wrote pseudotime, state and other factors to CSV, writting DE Genes for pseudotime")
    # sig_fit_coefs$term <- gsub("cluster", "", sig_fit_coefs$term)
    # sig_fit_coefs$term <- gsub("\\(Intercept\\)", "1", sig_fit_coefs$term)
    # DEGfile <- paste0(outPath, "/DEG_Clusters_", RunID, ".csv")
    # DEG <- subset(sig_fit_coefs, status=="OK")
    # DEG <- apply(DEG,2,as.character)
    # # write.csv(DEG, DEGfile, row.names = FALSE)
    #
    # write.table("[GeneSetCollection]", file = DEGfile, append = F, quote = F, row.names = F, col.names = F)
    # write.table("$SchemaVersion,1", file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
    # write.table(paste0("$GeneSetCollectionName,", "Monocle_Genes_",RunID), file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
    # write.table("", file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
    #
    # i=1
    #
    # for (ClustNum in unique(fit_coefs$term)){
    #     write.table("[GeneSet]", file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
    #     write.table(paste0("$GeneSetName,","Genes_",ClustNum), file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
    #     write.table("$GeneSetDescription,MadeInMonocle3", file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
    #     write.table(" ", file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
    #     write.table("[Genes], q_value", file = DEGfile, append = T, quote = F, row.names = F, col.names = F)
    #     write.table(DEG[DEG[,"term"] == i, c("gene_short_name", "q_value")], file = DEGfile, sep = ',', append = T, quote = F, row.names = F, col.names = F)
    #     i = i+1
    # }





# With regression:
# gene_fits <- fit_models(cds, model_formula_str = "~embryo.time")
# fit_coefs <- coefficient_table(gene_fits)
# emb_time_terms <- fit_coefs %>% filter(term == "embryo.time")
# emb_time_terms <- emb_time_terms %>% mutate(q_value = p.adjust(p_value))
# sig_genes <- emb_time_terms %>% filter (q_value < 0.05) %>% pull(gene_short_name)
#
# # With graph autocorrelation:
# pr_test_res <- graph_test(cds,  neighbor_graph="principal_graph", cores=4)
# pr_deg_ids <- row.names(subset(pr_test_res, q_value < 0.05))

# pr_graph_test_res <- graph_test(cds, neighbor_graph="principal_graph", cores=2)
# pr_graph_test_res <- pr_graph_test_res[order(pr_graph_test_res$q_value, decreasing = FALSE),]
# pr_deg_ids <- row.names(subset(pr_graph_test_res, q_value < 0.05))
# nGenes <- length(pr_deg_ids)
# if (nGenes >20){
#
# pr_deg_ids <- pr_deg_ids[1:20]
# }



######### Make outputs
###### Create DEG Gene tables
# for (i in unique(fit_coefs$term)){
# DEG <- subset(sig_fit_coefs, term==i)
# DEGfile <- paste0(outPath, "/DEG_", i, ".csv")
# DEG <- apply(DEG,2,as.character)
# write.csv(DEG, DEGfile)
# }



#########    Make Violinplots
    # if (VPBool){
    #     plot_genes_violin(cds, group_cells_by="cluster", ncol=2) +
    #     theme(axis.text.x=element_text(angle=45, hjust=1))
    # }

writeLines(" Gather Pseudotime analysis for export")
umapXY <- as.data.frame(t(cds@principal_graph_aux@listData[["UMAP"]][["pr_graph_cell_proj_dist"]]))
colnames(umapXY) <- c(paste0("Monocle_UMAP_X_", RunID), paste0("Monocle_UMAP_Y_", RunID))



outState <- pData(cds)
#rownames(pseudoXY) <- cellAnnotations$Sample
writeLines(" Format data output and prepare for images")
colnames(outState)[which(names(outState) == "Sample")] <- "CellId"
colnames(outState)[which(names(outState) == "Size_Factor")] <- paste0("Monocle_Size_Factor_", RunID)
colnames(outState)[which(names(outState) == "num_genes_expressed")] <- paste0("Monocle_Genes_Expressed_", RunID)
colnames(outState)[which(names(outState) == "Pseudotime")] <- paste0("Monocle_Pseudotime_", RunID)
colnames(outState)[which(names(outState) == "State")] <- paste0("Monocle_State_", RunID)

writeLines(" Plot heatmap for all clusters")



writeLines(" Gather Pseudotime analysis for export")
pseudotime <- as.data.frame(cds@principal_graph_aux@listData[["UMAP"]][["pseudotime"]])
pseudotime <- do.call(data.frame,lapply(pseudotime, function(x) replace(x, is.infinite(x),0)))
colnames(pseudotime) <- c(paste0("Monocle_Pseudotime_", RunID))
clusters <- as.data.frame((cds@clusters$UMAP$clusters))
colnames(clusters) <- c(paste0("Monocle_Clusters_", RunID))


CellIDCol <- colnames(cds)
CellIDCol <- gsub("AppID_*", "", CellIDCol)
outFile = cbind(CellIDCol, clusters, pseudotime, umapXY[,1:2])
colnames(outFile)[1] <- "CellID"

#write.csv(outFile, file = paste0(outPath, "Monocle_Analysis.csv"), quote = FALSE, row.names = FALSE)
writeLines(" Wrote pseudotime, state and other factors to CSV, writting DE Genes for pseudotime")

write.csv(outFile, file = paste0(outPath, "Monocle_R_Results.csv"), quote = FALSE, row.names = FALSE)





writeLines(" Finished plugin Monocle, returning to SeqGeq.")


